import logging
from datetime import datetime

import requests
from pydantic import validate_call

from .core import SanetId, sanetapi
from .schemas.models import NodesDict
from .schemas.outs import (
    AlarmsOut,
    ConditionOut,
    DevicesOut,
    InterfaceOut,
    InterfacesOut,
    NodeOut,
    NodesOut,
    PrimaryConditionStatusChangeOut,
    ServicesOut,
    StoragesOut,
    TagsOut,
    TagTreesOut,
    TrafficInterfaceOut,
)
from .settings import SanetSettings

logger = logging.getLogger(__name__)


class SanetApi:
    def __init__(self, settings: SanetSettings):
        self.settings = settings

    @property
    def auth_token_parameter(self):
        return "sanet3auth"

    @property
    def auth_data(self):
        return {self.auth_token_parameter: self.settings.auth_token.get_secret_value()}

    def call_api(
        self,
        method: str,
        path: str,
        params: dict | None = None,
        headers: dict | None = None,
        data: dict | str | None = None,
        **kwargs,
    ):
        url = f"{self.settings.api_url}{path}"
        request_method = getattr(requests, method)
        response = request_method(
            url, headers=headers, data=data, params=params, **kwargs
        )

        return response

    @sanetapi(schema_out=NodesOut)
    def read_all_nodes(
        self,
        **kwargs,
    ) -> NodesOut:
        result = self.call_api("get", "nodes", **kwargs)
        return result

    @validate_call
    @sanetapi(schema_out=NodeOut)
    def read_node(
        self,
        node_id: SanetId,
        **kwargs,
    ) -> NodeOut:
        result = self.call_api("get", f"node/{node_id}", **kwargs)
        return result

    @validate_call
    @sanetapi(schema_out=PrimaryConditionStatusChangeOut)
    def read_primary_condition_status_change(
        self,
        node_id: SanetId,
        start: datetime | None = None,
        end: datetime | None = None,
        **kwargs,
    ) -> PrimaryConditionStatusChangeOut:
        if start:
            kwargs["params"].update({"ts_start": int(start.timestamp())})
        if end:
            kwargs["params"].update({"ts_end": int(end.timestamp())})
        result = self.call_api(
            "get",
            f"node/{node_id}/get_primary_condition_status_change",
            **kwargs,
        )
        return result

    @sanetapi(schema_out=AlarmsOut)
    def read_all_alarms(
        self,
        priority_level: str | None = None,
        **kwargs,
    ) -> AlarmsOut:
        result = self.call_api("get", "alarm/current", **kwargs)
        return result

    @sanetapi(schema_out=InterfacesOut)
    def read_all_interfaces(
        self,
        **kwargs,
    ) -> InterfacesOut:
        result = self.call_api("get", "interfaces", **kwargs)
        return result

    @validate_call
    @sanetapi(schema_out=InterfaceOut)
    def read_interface(
        self,
        interface_id: SanetId,
        **kwargs,
    ) -> InterfaceOut:
        result = self.call_api("get", f"interface/{interface_id}", **kwargs)
        return result

    @validate_call
    @sanetapi(schema_out=TrafficInterfaceOut)
    def read_traffic_interface(
        self,
        interface_id: SanetId,
        ts_start: int | None = None,
        ts_end: int | None = None,
        **kwargs,
    ) -> TrafficInterfaceOut:
        result = self.call_api("get", f"interface/{interface_id}/get_traffic", **kwargs)
        return result

    @sanetapi(schema_out=StoragesOut)
    def read_all_storages(
        self,
        **kwargs,
    ) -> StoragesOut:
        result = self.call_api("get", "storages", **kwargs)
        return result

    @sanetapi(schema_out=ServicesOut)
    def read_all_services(
        self,
        **kwargs,
    ) -> ServicesOut:
        result = self.call_api("get", "services", **kwargs)
        return result

    @sanetapi(schema_out=DevicesOut)
    def read_all_devices(
        self,
        **kwargs,
    ) -> DevicesOut:
        result = self.call_api("get", "devices", **kwargs)
        return result

    @sanetapi(schema_out=ConditionOut)
    def read_all_conditions(
        self,
        **kwargs,
    ) -> ConditionOut:
        result = self.call_api("get", "conditions", **kwargs)
        return result

    @sanetapi(schema_out=TagTreesOut)
    def read_tagtrees(
        self,
        **kwargs,
    ) -> TagTreesOut:
        result = self.call_api("get", "tagtrees", **kwargs)
        return result

    @sanetapi(schema_out=TagsOut)
    def read_all_tags(
        self,
        **kwargs,
    ) -> TagsOut:
        result = self.call_api("get", "tags", **kwargs)
        return result


def read_all_nodes(api: SanetApi, key: str = "name") -> NodesDict:
    results = api.read_all_nodes()
    return NodesDict.create(nodes=results.result, key=key)



from datetime import datetime
from typing import Any

from pydantic import (
    UUID4,
    BaseModel,
    ConfigDict,
    RootModel,
    computed_field,
    validate_call,
)
from pydantic.networks import IPvAnyAddress

from ..enums import ElementType, PriorityLevel


class SanetBase(BaseModel):
    name: str | None = None
    path: str | None = None
    uuid: UUID4 | None = None


class SanetModel(SanetBase):
    cluster_distinguisher: str | None = None
    cluster_element_id: UUID4 | None = None
    cluster_instance: str | None = None
    cluster_instance_time: datetime | None = None  # TODO: verificare
    cluster_role: int | None = None
    cluster_xform: str | None = None
    cluster_id: UUID4 | None = None
    description: str | None = None
    kind: int | None = None
    parent_id: str | None = None


class SanetComponentModel(SanetModel):
    distinguisher: str | None = None
    is_abstract: bool | None = None
    xform: str | None = None


class Node(SanetModel):
    model_config = ConfigDict(extra="forbid")

    agent_id: UUID4 | None = None
    dft_timeout: int | None = None
    dns6_name: str | None = None
    dns_name: str | None = None
    effective_ip: IPvAnyAddress | None = None
    ip4_mgt: str | None = None
    ip6_mgt: str | None = None
    is_abstract: bool = False
    snmp_community: str | None = None
    snmp_port: int | None = None
    snmp_v3_authdata: str | None = None
    snmp_version: int | None = None
    tenant_id: UUID4 | None = None
    timezone: str | None = None
    use_ipv4: bool = True


class NodesDict(RootModel):
    root: dict[str, Node] = {}

    def __iter__(self):
        return iter(self.root)

    @validate_call
    def __getitem__(self, item: str) -> Node:
        return self.root[item]

    @validate_call
    def __contains__(self, key) -> bool:
        return key in self.root

    @validate_call
    def __setitem__(self, item: str, value: Node):
        self.root[item] = value

    def keys(self):
        return self.root.keys()

    @classmethod
    def create(cls, nodes: list[Node], key: str = "name") -> "NodesDict":
        data = {}
        for node in nodes:
            data[str(getattr(node, key))] = node
        return cls(data)


class Interface(SanetComponentModel):
    model_config = ConfigDict(extra="forbid")

    backbone: bool | None = None
    ifindex: str | None = None
    speed: float | None = None


class Storage(SanetComponentModel):
    model_config = ConfigDict(extra="forbid")

    stindex: str | None = None


class Service(SanetComponentModel):
    model_config = ConfigDict(extra="forbid")

    swrunindex: str | None = None


class Device(SanetComponentModel):
    model_config = ConfigDict(extra="forbid")

    devindex: str | None = None


class Element(SanetBase):
    model_config = ConfigDict(extra="forbid", validate_assignment=True)

    _base_url: str = ""

    description: str | None = None
    parent_id: UUID4 | None = None
    parent_name: str | None = None
    # *tipo dell'elemento (1=nodo,2=interfaccia,3=storage,4=service,5=device)*
    type: ElementType | None = None

    @computed_field  # type: ignore[misc]
    def absolute_url(self) -> str:
        return f"{self.type.name.lower()}/{self.uuid.hex}"

    @computed_field  # type: ignore[misc]
    def absolute_parent_url(self) -> str:
        return (
            f"{ElementType.NODE.name.lower()}/{self.parent_id.hex}"
            if self.parent_id
            else ""
        )

    def is_node(self):
        return self.type == ElementType.NODE


class ConditionBase(SanetBase):
    priority: int | None = None
    classification: str | None = None
    statuslastchange: datetime | None = None


class Condition(ConditionBase):
    model_config = ConfigDict(extra="forbid")

    statuschange_action: str | None = None
    uncheckable_fallback: int | None = None
    primary: bool | None = None
    mtime: datetime | None = None
    statuslast: Any | None = None
    lastchecktotaltime: Any | None = None  # TODO: verificare
    title: str | None = None
    element_id: UUID4 | None = None
    status: int | None = None
    tries: int | None = None
    passive_force_status: str | None = None
    max_tries: int | None = None
    dependson_id: str | None = None
    expr: str | None = None
    cascade: bool | None = None
    datagroup_id: UUID4 | None = None
    lastdone: int | None = None
    order: int | None = None


class AlarmCondition(ConditionBase):
    model_config = ConfigDict(extra="forbid")

    priority_level: PriorityLevel | None = None
    description: str | None = None


class Result(BaseModel):
    model_config = ConfigDict(extra="forbid")

    status: int | None = None
    laststatus: int | None = None


class Cluster(BaseModel):
    model_config = ConfigDict(extra="allow")

    name: list[str] | str | None = None


class Alarm(BaseModel):
    model_config = ConfigDict(extra="allow")

    info: dict | None = None
    datasources: dict | None = None
    tags: list[str] | None = None
    element: Element | None = None
    meta: dict | None = None
    result: Result | None = None
    condition: AlarmCondition | None = None
    cluster: Cluster | None = None


class TagBase(SanetBase):
    model_config = ConfigDict(extra="forbid")


class Tag(TagBase):
    model_config = ConfigDict(extra="allow")

    icon: str | None = None
    mtime: datetime | None = None
    parent_id: UUID4 | None = None
    tree_id: UUID4 | None = None


class TrafficInfo(BaseModel):
    in_avg: float | None = None
    in_max: float | None = None
    out_avg: float | None = None
    out_max: float | None = None



from pathlib import Path

import rich
import typer
from pydantic_settings_toml.exceptions import TomlSettingsError

from ..api import SanetApi
from ..exceptions import SanetError
from ..schemas.models import Element
from ..settings import AppSettings, SanetSettings


def init_app_settings(env_file: Path | None = None) -> AppSettings:
    try:
        return AppSettings(_env_file=env_file)
    except TomlSettingsError as tse:
        raise SanetError(tse) from tse


def cli_init_setting(env_file: Path | None = None) -> AppSettings:
    try:
        return init_app_settings(env_file=env_file)
    except SanetError as ie:
        rich.print(f"[bold red]{ie}[/bold red]")
        raise typer.Exit(code=10) from ie


def cli_init_api(env_file: Path | None = None) -> tuple[SanetApi, AppSettings]:
    settings = cli_init_setting(env_file=env_file)
    try:
        api = SanetApi(settings=settings.pysanet)
        return api, settings
    except Exception as exc:
        rich.print(f"[bold red]{exc}[/bold red]")
        raise typer.Exit(code=11) from exc


def get_element_descripion(setting: SanetSettings, element: Element):
    if element.is_node():
        return (
            f"[magenta2][link={setting.portal_url}{element.absolute_url}]{element.name}"
            f"[/link][/magenta2]"
        )

    return (
        f"[magenta2][link={setting.portal_url}{element.absolute_parent_url}]{element.parent_name}"
        f"[/link][/magenta2]"
        f":[cyan1][link={setting.portal_url}{element.absolute_url}]{element.name}[/link][/cyan1]"
    )



import functools
from typing import Annotated
from uuid import UUID

from pydantic import UUID4
from pydantic.functional_validators import AfterValidator


def sanetapi(func=None, *, schema_out=None):
    if func is None:
        return functools.partial(sanetapi, schema_out=schema_out)

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if "params" not in kwargs:
            kwargs["params"] = self.auth_data
        else:
            kwargs["params"].update(self.auth_data)
        for sanet_param in ["tenant", "username"]:
            if sanet_param in kwargs:
                kwargs["params"].update({sanet_param: kwargs[sanet_param]})

        response = func(self, *args, **kwargs)

        if not schema_out:
            return response
        result = schema_out.model_validate_json(response.content)
        result._response = response
        return result

    return wrapper


def check_sanet_id(value: str | UUID4) -> str:
    sanet_id = value.hex if isinstance(value, UUID) else value
    return sanet_id


SanetId = Annotated[str | UUID4, AfterValidator(check_sanet_id)]



from datetime import time
from typing import Any

from pydantic import UUID4, BaseModel, RootModel

from .models import (
    Alarm,
    Condition,
    Device,
    Interface,
    Node,
    Service,
    Storage,
    Tag,
    TagBase,
    TrafficInfo,
)


class SanetOut(BaseModel):
    elapsed_time: time | None = None
    operation: str | None = None
    result: list[Any] | None = None


class NodesOut(SanetOut):
    result: list[Node] | None = None


class NodeOut(SanetOut):
    result: Node | None = None


class InterfaceOut(SanetOut):
    result: Interface | None = None


class InterfacesOut(SanetOut):
    result: list[Interface] | None = None


class AlarmsOut(RootModel):
    root: list[Alarm] | None = None

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)


class PrimaryConditionStatusChangeOut(SanetOut):
    result: list[tuple] | None = None
    condition_path: str | None = None
    element_name: str | None = None
    element_path: str | None = None
    element_uuid: UUID4 | None = None


class TrafficInterfaceOut(SanetOut):
    result: TrafficInfo | None = None
    element_name: str | None = None
    element_path: str | None = None
    element_uuid: UUID4 | None = None


class StoragesOut(SanetOut):
    result: list[Storage] | None = None


class ServicesOut(SanetOut):
    result: list[Service] | None = None


class DevicesOut(SanetOut):
    result: list[Device] | None = None


class ConditionOut(SanetOut):
    result: list[Condition] | None = None


class TagTreesOut(SanetOut):
    result: list[TagBase] | None = None


class TagsOut(SanetOut):
    result: list[Tag] | None = None



from pathlib import Path
from zoneinfo import ZoneInfo

import rich  # noqa
import typer
from devtools import debug  # noqa
from rich.progress import Progress
from rich.table import Table

from .utils import cli_init_api, get_element_descripion

app = typer.Typer()
nodes_app = typer.Typer()
app.add_typer(nodes_app, name="nodes")
alarms_app = typer.Typer()
app.add_typer(alarms_app, name="alarms")

state = {"env_file": None}


@app.callback()
def set_env_file(
    env_file: Path = typer.Option(  # noqa: B008
        Path("~/.pynta.toml"), "--env", "-e", help="Environment file in formato toml"
    ),
):
    state["env_file"] = env_file


@alarms_app.command("list")
def list_alarms(
    node_name: str = typer.Option("", "--node-name", "-n"),  # noqa: B008
):
    api, _ = cli_init_api(env_file=state["env_file"])
    node_name = node_name.lower()

    with Progress() as progress:
        task = progress.add_task("Get data from Sanet...", total=None)
        alarms = api.read_all_alarms()
        progress.remove_task(task)

    if len(alarms) == 0:
        typer.Exit("...")

    table = Table("Element", "Condition", "Priority", "Last Change", show_lines=True)

    if node_name:
        alarms = [alarm for alarm in alarms if node_name in alarm.element.path.lower()]

    for alarm in alarms:
        table.add_row(
            get_element_descripion(setting=api.settings, element=alarm.element),
            alarm.condition.description,
            alarm.condition.priority_level,
            alarm.condition.statuslastchange.astimezone(
                tz=ZoneInfo("Europe/Rome")
            ).strftime("%Y-%m-%d %H:%M"),
        )
    rich.print(table)
